import { TypeTracker } from "@previewjs/ts-helpers";
import {
  createFileSystemReader,
  createMemoryReader,
  createStackedReader,
  Reader,
  Writer,
} from "@previewjs/vfs";
import path from "path";
import { extractReactComponents } from "./extract-component";

const MAIN_FILE = path.join(__dirname, "virtual", "App.tsx");

describe("extractReactComponents", () => {
  let memoryReader: Reader & Writer;
  let typeTracker: TypeTracker;

  beforeEach(() => {
    memoryReader = createMemoryReader();
    typeTracker = new TypeTracker({
      rootDirPath: path.join(__dirname, "virtual"),
      reader: createStackedReader([
        memoryReader,
        createFileSystemReader({
          mapping: {
            from: path.resolve(__dirname, "..", "..", "..", "node_modules"),
            to: path.join(__dirname, "virtual", "node_modules"),
          },
        }),
        createFileSystemReader(), // required for TypeScript libs, e.g. Promise
      ]),
      typeRoots: [],
    });
    typeTracker.start();
  });

  afterEach(() => {
    typeTracker.stop();
  });

  it("detects expected components", async () => {
    expect(
      extract(`
import React, { Component } from "react";

function wrapComponent<P>(c: React.ComponentType<P>): React.ComponentType<P> {
  return c;
}

function DeclaredFunction() {
  return <div>Hello, World!</div>;
}

const ConstantFunction = () => <div>Hello, World!</div>;

// Note: this isn't detected as of October 2021.
const HocComponent = wrapComponent(() => <div>Hello, World!</div>);

class ClassComponent1 extends React.Component {}

class ClassComponent2 extends Component {}

const ForwardRef = React.forwardRef((props, ref) => <ConstantFunction {...props} />);

export class BaseClass<T> {}

export class NotComponentClass extends BaseClass<SomeProps> {}

export const NotObjectProps = (props: () => void) => {
  return <div>Hello, World!</div>;
};

export const MissingType = (props) => {
  return <div>Hello, World!</div>;
};

export const NotComponent = () => {
  // This isn't a component.
};

export const NotComponentEither = () => {
  return "Hello";
};
`)
    ).toEqual([
      {
        name: "DeclaredFunction",
        exported: false,
        offsets: expect.anything(),
      },
      {
        name: "ConstantFunction",
        exported: false,
        offsets: expect.anything(),
      },
      // Note: this isn't detected as of October 2021.
      // {
      //   name: "HocComponent",
      //   exported: false,
      //   offsets: expect.anything(),
      // },
      {
        name: "ClassComponent1",
        exported: false,
        offsets: expect.anything(),
      },
      {
        name: "ClassComponent2",
        exported: false,
        offsets: expect.anything(),
      },
      {
        name: "ForwardRef",
        exported: false,
        offsets: expect.anything(),
      },
      {
        name: "NotObjectProps",
        exported: true,
        offsets: expect.anything(),
      },
      {
        name: "MissingType",
        exported: true,
        offsets: expect.anything(),
      },
    ]);
  });

  function extract(source: string) {
    const rootDirPath = path.join(__dirname, "virtual");
    memoryReader.updateFile(path.join(rootDirPath, "App.tsx"), source);
    return extractReactComponents(
      typeTracker.analyzeProgram([MAIN_FILE]),
      MAIN_FILE
    );
  }
});
